const express = require('express')
const serumsRoutes = require('./src/api/REST/serums/serumsRoutes')()
const noCache = require('./src/utils/express/middleware/noCache')
const { CHAINCODE_ERROR } = require('./src/constants/errorKeys')

const logger = require('./src/utils/logging/logger').getLogger('initApi')

module.exports = function initApi (app, apiBasePath) {
  const apiRouter = express.Router()

  // Middleware
  noCache(apiRouter)

  app.use(apiBasePath, apiRouter)

  apiRouter.use('/', serumsRoutes)

  const API_ERRORS = ['NOT_AUTHENTICATED', 'INVALID_TOKEN', 'NOT_AUTHORIZED', 'NO_FILTER']
  const AUTH_MODULE_ERRORS = ['AUTHENTICATION_ERROR']

  function errorOrigin (key) {
    if (API_ERRORS.includes(key)) {
      return 'BLOCKCHAIN_API'
    }
    if (AUTH_MODULE_ERRORS.includes(key)) {
      return 'AUTHENTICATION_MODULE'
    }

    return 'BLOCKCHAIN_HYPERLEDGER'
  }

  function errorMessage (key, message) {
    if (API_ERRORS.includes(key) || AUTH_MODULE_ERRORS.includes(key)) {
      return message
    }

    const regex = /(?<="message":")(.*?)(?="\})/
    const match = message.match(regex)

    if (match == null) {
      return `HYPERLEDGER ERROR: ${message}`
    }

    return `HYPERLEDGER ERROR: ${match[0]}`
  }

  // eslint-disable-next-line no-unused-vars
  app.use((error, req, res, next) => {
    if (error.status && error.key) {
      logger.info('Api request failed', {
        status: error.status,
        key: error.key,
        error: JSON.stringify(error.message)
      })
    } else {
      logger.error('Api request failed with an unexpected error', {
        error: error.message
      })
    }

    if (error.key === CHAINCODE_ERROR) {
      res.status(500)
    } else {
      res.status(error.status || 500)
    }

    return res.json({
      origin: errorOrigin(error.key),
      message: errorMessage(error.key, error.message),
      error
    })
  })
}
