const { network } = require('../network')

const CONTRACT_NAME = 'serums'

function isString (value) {
  return typeof value === 'string' || value instanceof String
}

function createSerumsContract (config) {
  async function getContract (org) {
    const orgConfig = config[org]
    if (!orgConfig) {
      throw new Error(`No blockchain configured for organisation ${org}`)
    }

    const {
      chaincodeUser,
      channel
    } = orgConfig

    return network.organizations[orgConfig.orgId].contractFor(
      CONTRACT_NAME,
      chaincodeUser, // user
      channel// 'channel'
    )
  }

  async function query (fct, org, ...args) {
    const [contract, gateway] = await getContract(org)

    // Query
    const result = await contract.evaluateTransaction(
      fct,
      ...args.filter((arg) => arg).map((arg) => (isString(arg) ? arg : JSON.stringify(arg)))
    )
    gateway.disconnect()

    const resultObject = JSON.parse(result.toString())
    return resultObject
  }

  async function invoke (fct, org, ...args) {
    const [contract, gateway] = await getContract(org)
    // Invoke
    const result = await contract.submitTransaction(
      fct,
      ...args.filter((arg) => arg).map((arg) => (isString(arg) ? arg : JSON.stringify(arg)))
    )
    gateway.disconnect()

    const resultObject = JSON.parse(result.toString())
    return resultObject
  }

  async function createRule (org, jwtToken, data) {
    return invoke('createRule', org, jwtToken, data)
  }

  async function updateRule (org, jwtToken, data, ruleId) {
    return invoke('updateRule', org, jwtToken, data, ruleId)
  }

  async function deleteRule (org, jwtToken, data) {
    return invoke('deleteRule', org, jwtToken, data)
  }

  async function getRules (org, data) {
    return query('getRules', org, data)
  }

  async function getPatientAccess (org, jwtToken, data) {
    return invoke('getPatientAccess', org, jwtToken, data)
  }

  async function getPatientAccessAll (org, jwtToken, data) {
    return invoke('getPatientAccessAll', org, jwtToken, data)
  }

  async function getPatientAccessRead (org, jwtToken, data) {
    return query('getPatientAccess', org, jwtToken, data)
  }

  async function getPatientAccessAllRead (org, jwtToken, data) {
    return query('getPatientAccessAll', org, jwtToken, data)
  }

  async function getAccessLog (org, data) {
    return query('getAccessLog', org, data)
  }

  async function ping (org, data) {
    return query('ping', org, data)
  }

  async function reset (org) {
    return invoke('reset', org)
  }

  async function createProof (org, data) {
    return invoke('createProof', org, data)
  }

  async function updateProof (org, data, proofId) {
    return invoke('updateProof', org, data, proofId)
  }

  async function getProof (org, proofId) {
    return query('getProof', org, proofId)
  }

  return {
    createRule,
    updateRule,
    deleteRule,
    getRules,
    getPatientAccess,
    getPatientAccessAll,
    getPatientAccessRead,
    getPatientAccessAllRead,
    getAccessLog,
    ping,
    reset,
    createProof,
    updateProof,
    getProof
  }
}

module.exports = createSerumsContract
