const FabricClient = require('fabric-client')
const {Gateway/* , Network, Contract, ContractEventListener */} = require('fabric-network')

const networkConfig = require('../../config/network')
const {IdentityService} = require('./identityService')
// import { IListenerOptions, IConnectOptions } from './typings';
// import { INetworkConfig } from '../config/typings';
// import { OrgType } from '../config/orgTypings';

// interface IEvent {
//     // eslint-disable-next-line camelcase
//     chaincode_id: string;
//     // eslint-disable-next-line camelcase
//     tx_id: string;
//     // eslint-disable-next-line camelcase
//     event_name: string;
//     payload: Buffer;
// }

class Organization {
    // name = undefined // String

    // identityService = undefined // IdentityService

    constructor(name) {
        this.name = name
        this.identityService = new IdentityService(this.name, this.getClient())
    }

    async listenerFor(
        contractName, // string
        listenerName, // string
        eventName, // string
        // callback is a function: (
        //     error, // Error
        //     payload, // object|string|null
        //     blockNumber, // string
        //     txId, // string
        //     status //string
        // ) => Promise<void>
        callback,
        user = 'admin', // string
        channel = 'default', // string
        options = {}, // IListenerOptions
    ) { // Returns Promise<ContractEventListener>
        const [contract] = await this.contractFor(
            contractName,
            user,
            channel,
            {checkpointer: options.checkpointer}
        )

        const listener = await contract.addContractListener(
            `${this.name}/${listenerName}`,
            // is being put in a regexp via ChaincodeRegistration ... and else publish and unpublish event will match
            // the full name should match
            `^${eventName}$`,
            // eslint-disable-next-line max-len
            async(
                error, // Error
                event, // IEvent
                blockNumber, // string
                txId, // string
                status, // string
            ) => { // Returns Promise<void>
                // Extract and parse the payload
                const {payload: rawPayload} = event
                let payload = null
                if (rawPayload && Buffer.isBuffer(rawPayload)) {
                    payload = rawPayload.toString('utf8')
                } else if (typeof rawPayload === 'string') {
                    payload = rawPayload
                }

                if (typeof payload === 'string') {
                    try {
                        // try parsing it
                        payload = JSON.parse(payload)
                    } catch (e) { /* do nothing */ }
                }

                callback(error, payload, blockNumber, txId, status)
            },
            {
                ...options,
                replay: (options.checkpointer != null)
            },
        )

        return listener
    }

    async contractFor(
        contractName, // string,
        user = 'admin', // string
        channel = 'default', // string
        options = {}, // IConnectOptions
    ) { // Returns Promise<[Contract, Gateway, Network]>
        const [network, gateway] = await this.networkFor(user, channel, options)
        const contract = network.getContract(contractName)

        return [contract, gateway, network]
    }

    async networkFor(
        user = 'admin', // string,
        channel = 'default', // string,
        options = {}, // IConnectOptions,
    ) { // Returns Promise<[Network, Gateway]>
        const gateway = await this.gatewayFor(user, options)
        const network = await gateway.getNetwork(channel)

        return [network, gateway]
    }

    async gatewayFor(
        user = 'admin', // string,
        options = {}, // IConnectOptions
    ) { // Returns Promise<Gateway>
        const client = this.getClient()
        const gateway = new Gateway()
        const wallet = await this.identityService.getWallet()

        await gateway.connect(client, {
            wallet,
            identity: user,
            discovery: {enabled: false},
            ...options
        })

        return gateway
    }

    // Private
    getClient() { // Returns FabricClient
        if (networkConfig.organizations[this.name] == null) {
            throw new Error(`Unknown organization ${this.name}.`)
        }

        const updatedConfig = { // INetworkConfig
            ...networkConfig,
            client: {
                organization: this.name,
                credentialStore: {
                    path: `${networkConfig.client.credentialStore.path}/${this.name}`,
                    cryptoStore: {
                        path: `${networkConfig.client.credentialStore.cryptoStore.path}/${this.name}`
                    }
                }
            }
        }

        return FabricClient.loadFromConfig(updatedConfig)
    }
}

const network = {
    organizations: Object.keys(networkConfig.organizations).reduce((acc, orgName) => {
        acc[orgName] = new Organization(orgName)

        return acc
    }, {})
}

module.exports = {
    Organization,
    network
}
