const {
    Transform
} = require('@theledger/fabric-mock-stub')

const {
    ruleIdRegex
} = require('../shapes')

const getToken = require('./getToken')
let txId = 0

class Rule {
    /**
    *  Usage:
    *       new Rule()
    *           .allow('DOCTOR_1')
    *           .from('HOSPITAL_1')
    *           .toAccess(['EYE', 'BRAIN'])
    *           .of('PATIENT1')
    *           .from('HOSPITAL_1')
    *
    *   Or:
    *       new Rule()
    *           .allowAll('DOCTOR')
    *           .from('HOSPITAL_1')
    *           .toAccess(['EYE', 'BRAIN'])
    *           .ofAllPatientsOf('HOSPITAL_1')
    *           .asUser('USER_1')
    */
    constructor() {
        this.rule = {
            grantor: {},
            grantee: {}
        }

        this.rule.access = []
        this.rule.expires = new Date('2025-01-30').toISOString()

        this.user = {
            groupIds: ['HOSPITAL_ADMIN']
        }
    }

    nextTxId() {
        txId += 1
        return `REQUEST_${txId}`
    }

    allow(userId) {
        this.rule.action = 'ALLOW'
        this.rule.grantee.type = 'INDIVIDUAL'
        this.rule.grantee.id = userId
        return this
    }

    deny(userId) {
        this.rule.action = 'DENY'
        this.rule.grantee.type = 'INDIVIDUAL'
        this.rule.grantee.id = userId
        return this
    }

    allowAll(groupId) {
        this.rule.action = 'ALLOW'
        this.rule.grantee.type = 'GROUP'
        this.rule.grantee.id = groupId
        return this
    }

    denyAll(groupId) {
        this.rule.action = 'DENY'
        this.rule.grantee.type = 'GROUP'
        this.rule.grantee.id = groupId
        return this
    }

    allowDepartment(departmentId) {
        this.rule.action = 'ALLOW'
        this.rule.grantee.type = 'DEPARTMENT'
        this.rule.grantee.id = departmentId
        return this
    }

    denyDepartment(departmentId) {
        this.rule.action = 'DENY'
        this.rule.grantee.type = 'DEPARTMENT'
        this.rule.grantee.id = departmentId
        return this
    }

    user(user) {
        this.rule.grantee.type = 'INDIVIDUAL'
        this.rule.grantee.id = user
        return this
    }

    of(patientId) {
        this.rule.grantor.type = 'INDIVIDUAL'
        this.rule.grantor.id = patientId
        this.user.id = patientId
        return this
    }

    from(orgId) {
        if (this.rule.grantee.orgId) {
            this.rule.grantor.orgId = orgId
            this.user.orgId = orgId
        } else {
            this.rule.grantee.orgId = orgId
        }
        return this
    }

    ofAllPatientsOf(orgId) {
        this.rule.grantor.orgId = orgId
        this.user.orgId = orgId
        this.rule.grantor.type = 'GROUP'
        return this
    }

    asUser(userId) {
        this.rule.grantor.id = userId
        this.user.id = userId
        return this
    }

    asAnotherUser(userId) {
        this.user.id = userId
        return this
    }

    ofAnotherOrganisation(orgId) {
        this.user.orgId = orgId
        return this
    }

    withGroups(groupIds) {
        this.user.groupIds = groupIds
        return this
    }

    makeExpired() {
        this.rule.expires = new Date('2015-01-30').toISOString()
        return this
    }

    toAccess(attrs) {
        this.rule.access = attrs.map((att) => ({name: att}))
        return this
    }

    getFullRule() {
        return this.rule
    }

    async createRule(mockStub, expectSuccess = true) {
        const ruleToCreate = {
            action: this.rule.action,
            grantor: {
                type: this.rule.grantor.type,
                id: this.rule.grantor.id
            },
            grantee: this.rule.grantee,
            access: this.rule.access,
            expires: this.rule.expires
        }

        const token = getToken({
            id: this.user.id,
            orgId: this.user.orgId,
            groupIds: this.user.groupIds,
            departmentIds: this.user.departmentIds
        })
        const response = await mockStub.mockInvoke(
            this.nextTxId(),
            ['createRule', token, JSON.stringify(ruleToCreate)]
        )

        if (expectSuccess) {
            expect(response.status).toBe(200)
            const payload = Transform.bufferToObject(response.payload)
            expect(payload.ruleId).toEqual(expect.stringMatching(ruleIdRegex))
            return payload.ruleId
        }
        expect(response.status).toBe(500)
        const error = Transform.bufferToObject(response.message)
        return error
    }

    async updateRule(mockStub, ruleId, expectSuccess = true) {
        const ruleToCreate = {
            action: this.rule.action,
            grantor: {
                type: this.rule.grantor.type,
                id: this.rule.grantor.id
            },
            grantee: this.rule.grantee,
            access: this.rule.access,
            expires: this.rule.expires
        }

        const token = getToken({
            id: this.user.id,
            orgId: this.user.orgId,
            groupIds: this.user.groupIds
        })

        const response = await mockStub.mockInvoke(
            this.nextTxId(),
            ['updateRule', token, JSON.stringify(ruleToCreate), ruleId]
        )

        if (expectSuccess) {
            expect(response.status).toBe(200)
            const payload = Transform.bufferToObject(response.payload)
            expect(payload.ruleId).toEqual(expect.stringMatching(ruleIdRegex))
            return payload.ruleId
        }
        expect(response.status).toBe(500)
        const error = Transform.bufferToObject(response.message)
        return error
    }
}

module.exports = Rule
