#!/bin/bash

set -e -o pipefail

usage() {
    echo "Usage: $(basename "$0") COMMAND"
    echo """commands:

    reset:
      Stops and destroys the blockchain. Does not delete the artifacts.

    init:
      Initializes the necessary artifacts, overrides existing ones.

    install:
      Installs and sets up the network.

    channels:
      Configures the channels.

    chaincodes [version]:
      If no version is specified, installs the chaincodes on the blockchain.
      If a version is specified, upgrades the chaincodes to that particular version.

    hosts:
      Once the blockchain is running, use this command to map the ips of the services of the kubernetes cluster to locally resolvable DNS names for each service (modifies /etc/hosts)

    ping:
      Test the ping command on the blockchain, if this succeeds it's a good sign all is running fine.

    list [installed|instantiated]:
      List installed or instantiated chaincodes.

    NOTE:
      In a normal flow, if you started from scratch, you would run init, install, channels, chaincodes, and hosts in this order.
"""
    exit 1
}

command=$1

INSTALL_DIR="$( cd "$( dirname "${BASH_SOURCE[0]}" )" && pwd )"

config=$INSTALL_DIR/config

export FABRIC_CFG_PATH=$config

hosts() {
	kubectl get svc -l addToHostAliases=true -o jsonpath='{range..items[*]}{.spec.clusterIP} {.metadata.labels.fqdn}{"\n"}{end}' | xargs -n 2 -L1 sh -c 'sudo sed -i.bak "/$2/d" /etc/hosts && sudo rm /etc/hosts.bak && sudo bash -c "echo $1 $2 >> /etc/hosts"' argv0
	echo "INFO: You need to export HOSTALIASES=$INSTALL_DIR/hlf-kube/hosts"
}

reset() {
	argo delete --all
	helm delete hlf-kube --purge
}

setup_env() {
	export FABRIC_CFG_PATH=$config
        export CRYPTO_CONFIG=$INSTALL_DIR/hlf-kube/crypto-config

        export CORE_PEER_MSPCONFIGPATH="$CRYPTO_CONFIG/peerOrganizations/serums1.com/users/Admin@serums1.com/msp"
        export CORE_PEER_ID=hlf-peer--org1--peer0
        export CORE_PEER_LOCALMSPID=Org1MSP
        export CORE_PEER_ADDRESS=peer0.serums1.com:7051
	export ORDERER_CERT=$CRYPTO_CONFIG/ordererOrganizations/serums.com/tlsca/tlsca.serums_com-cert.pem
	export ORDERER_ADDR=orderer.serums.com:7050
}

list() {
	setup_env
	instantiated_or_installed=$1
	if [ "$instantiated_or_installed" == "instantiated" ]; then
		peer chaincode --cafile $ORDERER_CERT --orderer $ORDERER_ADDR list --instantiated --channelID common
	elif [ "$instantiated_or_installed" == "installed" ]; then
                peer chaincode --cafile $ORDERER_CERT --orderer $ORDERER_ADDR list --installed --channelID common
	else
		echo "Must specify installed or instantiated"
		exit 1
	fi
}

ping() {
	setup_env
	peer chaincode --cafile $ORDERER_CERT --orderer $ORDERER_ADDR invoke --name "serums" --ctor '{"function":"ping","Args":[]}' --channelID common
}

init() {
	./init.sh $config $INSTALL_DIR/chaincode/
}

install() {
        echo "I AM HERE"
				# Alternative Helm version
				helm install hlf-kube $INSTALL_DIR/hlf-kube -f $config/network.yaml -f $config/crypto-config.yaml
				# helm install --name hlf-kube $INSTALL_DIR/hlf-kube -f $config/network.yaml -f $config/crypto-config.yaml
        echo "I AM DONE"
}

channels() {
        echo "CHANNELS"
				helm template $INSTALL_DIR/flows/channel-flow/ -f $config/network.yaml -f $config/crypto-config.yaml | argo submit - --watch
        #helm template $INSTALL_DIR/channel-flow/ -f $config/network.yaml -f $config/crypto-config.yaml | argo submit --watch
}

chaincodes() {
	./prepare_chaincodes.sh $config $INSTALL_DIR/chaincode
	helm upgrade hlf-kube $INSTALL_DIR/hlf-kube -f $config/network.yaml -f $config/crypto-config.yaml
	version=$1
        if [ -z "$version" ]; then
		echo "installing chaincode"
                helm template $INSTALL_DIR/flows/chaincode-flow/ -f $config/network.yaml -f $config/crypto-config.yaml | argo submit - --watch
                #helm template $INSTALL_DIR/chaincode-flow/ -f $config/network.yaml -f $config/crypto-config.yaml | argo submit --watch
	else
		echo "upgrading chaincode to $version"
		helm template $INSTALL_DIR/flows/chaincode-flow/ -f $config/network.yaml -f $config/crypto-config.yaml -f $INSTALL_DIR/flows/chaincode-flow/values.upgrade.yaml --set chaincode.version=$version | argo submit - --watch
	fi
}

if [ "$command" == "reset" ]; then
	reset
elif [ "$command" == "init" ]; then
	init
elif [ "$command" == "install" ]; then
	install
elif [ "$command" == "channels" ]; then
	channels
elif [ "$command" == "chaincodes" ]; then
	chaincodes $2
elif [ "$command" == "hosts" ]; then
	hosts
elif [ "$command" == "ping" ]; then
	ping
elif [ "$command" == "list" ]; then
	list $2
else
	echo "invalid command $command"
	usage
fi
